module HAL where

{-# LANGUAGE BangPatterns #-}

import Conduit
import Control.Monad
import Control.Monad.Reader
import Data.Aeson
import Data.LanguageCodes (ISO639_1(..))
import Data.Text
import HAL.Client
import HAL.Doc.Corpus
import HAL.Doc.Struct
import HAL.Utils (langTitleS, langAbstractS, toText)
import Network.HTTP.Client (newManager, Request)
import Network.HTTP.Client.TLS (tlsManagerSettings)
import Protolude
import Servant.API
import Servant.Client (BaseUrl(..), Scheme(..), ClientM, ClientError, runClientM, mkClientEnv, ClientEnv (makeClientRequest))
import System.IO.Unsafe (unsafePerformIO)

data HalCrawlerOptions
  = HalCrawlerOptions
  { -- | If 'True', enable the debug logs to stdout.
    _hco_debugLogs :: !Bool
  , _hco_batchSize :: !Int
  }

defaultHalOptions :: HalCrawlerOptions
defaultHalOptions = HalCrawlerOptions
  { _hco_debugLogs = False
  , _hco_batchSize = 1000
  }

type Query = Text
type Start = Int
type Limit = Integer
type Count = Integer

queryWithLang :: Maybe ISO639_1 -> [Query] -> [Query]
queryWithLang Nothing qs = qs
queryWithLang (Just lang) qs = qs <> ["language_s:" <> toText lang]

getMetadataWith :: [Query]
                 -- ^ The textual query
                -> Maybe Start
                 -- ^ An optional offset
                -> Maybe Limit
                -- ^ An optional limit for the search, it influences the number of
                -- rows returned.
                -> Maybe ISO639_1
                -- ^ An optional language for the search.
                -> IO (Either ClientError (Response Corpus))
getMetadataWith qs start_ limit lang = do
  runHalAPIClient defaultHalOptions $ search (Just $ requestedFields lang) (queryWithLang lang qs) Nothing start_ (fromIntegral  <$> limit)

getMetadataWithC :: [Query]
                 -- ^ The textual query
                 -> Maybe Start
                 -- ^ An optional offset
                 -> Maybe Limit
                 -- ^ An optional limit for the search, it influences the number of
                 -- rows returned.
                 -> Maybe ISO639_1
                 -- ^ An optional language for the search.
                 -> IO (Either ClientError (Maybe Count, ConduitT () Corpus IO ()))
getMetadataWithC qs start_ limit lang = getMetadataWithLangC defaultHalOptions (queryWithLang lang qs) start_ limit lang

getMetadataWithLangC :: HalCrawlerOptions
                     -- ^ The options for the crawler
                     -> [Query]
                     -- ^ The textual query
                     -> Maybe Start
                     -- ^ An optional offset
                     -> Maybe Limit
                     -- ^ An optional limit for the search, it influences the number of
                     -- rows returned.
                     -> Maybe ISO639_1
                     -- ^ An optional language for the search.
                     -> IO (Either ClientError (Maybe Count, ConduitT () Corpus IO ()))
getMetadataWithLangC opts@HalCrawlerOptions { .. } qs mb_offset mb_limit lang = do
  -- First, estimate the total number of documents
  eCount <- countResults qs
  pure $ get' <$> eCount
  where
    get' :: Count
         -> (Maybe Count, ConduitT () Corpus IO ())
    get' numFound' =
      ( Just numResults
      , yieldMany [0..]
        .| takeC (fromIntegral numPages)
        .| concatMapMC (getPage offset))
      where
        offset = fromMaybe 0 mb_offset
        limit  = min numFound' $ fromMaybe numFound' mb_limit
        numResults = limit - fromIntegral offset
        numPages = numResults `div` fromIntegral _hco_batchSize + 1

    getPage :: Start -> Int -> IO [Corpus]
    getPage start' pageNum = do
      let offset = start' + pageNum * _hco_batchSize
      debugLog opts $ "[getMetadataWithC] getPage: " <> show offset
      eRes <- runHalAPIClient opts $ search (Just $ requestedFields lang) qs Nothing (Just offset) (Just $ fromIntegral _hco_batchSize)
      pure $ case eRes of
        Left _ -> []
        Right Response { _docs } -> _docs

debugLog :: HalCrawlerOptions -> Text -> IO ()
debugLog HalCrawlerOptions{..} msg = when _hco_debugLogs $ putStrLn msg

countResults :: [Query] -> IO (Either ClientError Count)
countResults qs = do
  -- First, estimate the total number of documents
  eRes <- runHalAPIClient defaultHalOptions $ search (Just $ requestedFields Nothing) qs Nothing (Just 0) (Just 1) :: IO (Either ClientError (Response Corpus))
  pure $ fromIntegral <$> _numFound <$> eRes

requestedFields :: Maybe ISO639_1 -> Text
requestedFields (Just EN) = "docid,title_s,en_abstract_s,submittedDate_s,source_s,authFullName_s,authOrganism_s"
requestedFields (Just lang) = "docid," <> langTitleS lang <> "," <> langAbstractS lang <> ",submittedDate_s,source_s,authFullName_s,authOrganism_s"
requestedFields _ = requestedFields (Just EN)

structFields :: Text
structFields = "docid,label_s,parentDocid_i"

requestLog :: HalCrawlerOptions -> Request -> Request
requestLog opts rq = unsafePerformIO $ do
  debugLog opts $ "[HAL.makeClientRequestLog] " <> show rq
  pure rq
{-# NOINLINE requestLog #-}

runHalAPIClient :: (FromJSON doc, ToHttpApiData doc)
                => HalCrawlerOptions
                -> ClientM (Response doc)
                -> IO (Either ClientError (Response doc))
runHalAPIClient opts cmd = do
  manager' <- newManager tlsManagerSettings
  runClientM cmd' (mkClientEnv manager' $ BaseUrl Https "api.archives-ouvertes.fr" 443 "")
  where
    cmd' = local (\r -> r {
      makeClientRequest = \bUrl servantRq -> requestLog opts ((makeClientRequest r) bUrl servantRq)
      }) cmd

runStructureRequest :: Maybe Text -> IO (Either ClientError (Response Struct))
runStructureRequest rq =
  runHalAPIClient defaultHalOptions $ structure (Just structFields) rq (Just 10000)

runSearchRequest :: [Text] -> IO (Either ClientError (Response Corpus))
runSearchRequest rq =
  runHalAPIClient defaultHalOptions $ search (Just $ requestedFields Nothing) rq Nothing Nothing Nothing

generateRequestByStructID :: Text -> [Text] -> Text
generateRequestByStructID rq struct_ids =
  "text:" <>
  rq
  <> " AND "
  <> "structId_i:("
  <> (flattenPipe struct_ids)
  <> ")"

flattenPipe :: [Text] -> Text
flattenPipe [] = ""
flattenPipe (x:[]) = x
flattenPipe (x:xs) = x <> " || " <> flattenPipe xs
